package org.jastadd.ragconnect.tests;

import defaultOnlyRead.ast.A;
import defaultOnlyRead.ast.BoxedTypes;
import defaultOnlyRead.ast.MqttHandler;
import defaultOnlyRead.ast.NativeTypes;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.nio.file.Paths;
import java.util.concurrent.TimeUnit;

import static org.jastadd.ragconnect.tests.TestUtils.*;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Test case "defaultOnlyRead".
 *
 * @author rschoene - Initial contribution
 */
public class DefaultOnlyReadTest extends AbstractMqttTest {

  private static final String TOPIC_NATIVE_INT = "native/boolean";
  private static final String TOPIC_NATIVE_BOOLEAN = "native/int";
  private static final String TOPIC_NATIVE_SHORT = "native/short";
  private static final String TOPIC_NATIVE_LONG = "native/long";
  private static final String TOPIC_NATIVE_FLOAT = "native/float";
  private static final String TOPIC_NATIVE_DOUBLE = "native/double";
  private static final String TOPIC_NATIVE_CHAR = "native/char";
  private static final String TOPIC_NATIVE_STRING = "native/string";

  private static final String TOPIC_BOXED_BOOLEAN = "boxed/boolean";
  private static final String TOPIC_BOXED_INTEGER = "boxed/Integer";
  private static final String TOPIC_BOXED_SHORT = "boxed/Short";
  private static final String TOPIC_BOXED_LONG = "boxed/Long";
  private static final String TOPIC_BOXED_FLOAT = "boxed/Float";
  private static final String TOPIC_BOXED_DOUBLE = "boxed/Double";
  private static final String TOPIC_BOXED_CHARACTER = "boxed/Character";

  private A model;
  private NativeTypes integers;
  private NativeTypes floats;
  private NativeTypes chars;
  private BoxedTypes allBoxed;

  @Test
  public void checkNotJacksonReference() {
    testJaddContainReferenceToJackson(
        Paths.get("src", "test",
            "02-after-ragconnect", "defaultOnlyRead", "RagConnect.jadd"), false);
  }

  @Override
  protected void createModel() {
    model = new A();
    integers = new NativeTypes();
    model.addNativeTypes(integers);
    floats = new NativeTypes();
    model.addNativeTypes(floats);
    chars = new NativeTypes();
    model.addNativeTypes(chars);
    allBoxed = new BoxedTypes();
    model.addBoxedTypes(allBoxed);
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException {
    model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);

    integers.connectBooleanValue(mqttUri(TOPIC_NATIVE_BOOLEAN));
    integers.connectIntValue(mqttUri(TOPIC_NATIVE_INT));
    integers.connectShortValue(mqttUri(TOPIC_NATIVE_SHORT));
    integers.connectLongValue(mqttUri(TOPIC_NATIVE_LONG));
    floats.connectFloatValue(mqttUri(TOPIC_NATIVE_FLOAT));
    floats.connectDoubleValue(mqttUri(TOPIC_NATIVE_DOUBLE));
    chars.connectCharValue(mqttUri(TOPIC_NATIVE_CHAR));
    chars.connectStringValue(mqttUri(TOPIC_NATIVE_STRING));

    integers.connectBooleanValueTransformed(mqttUri(TOPIC_NATIVE_BOOLEAN));
    integers.connectIntValueTransformed(mqttUri(TOPIC_NATIVE_INT));
    integers.connectShortValueTransformed(mqttUri(TOPIC_NATIVE_SHORT));
    integers.connectLongValueTransformed(mqttUri(TOPIC_NATIVE_LONG));
    floats.connectFloatValueTransformed(mqttUri(TOPIC_NATIVE_FLOAT));
    floats.connectDoubleValueTransformed(mqttUri(TOPIC_NATIVE_DOUBLE));
    chars.connectCharValueTransformed(mqttUri(TOPIC_NATIVE_CHAR));
    chars.connectStringValueTransformed(mqttUri(TOPIC_NATIVE_STRING));

    allBoxed.connectBooleanValue(mqttUri(TOPIC_BOXED_BOOLEAN));
    allBoxed.connectIntValue(mqttUri(TOPIC_BOXED_INTEGER));
    allBoxed.connectShortValue(mqttUri(TOPIC_BOXED_SHORT));
    allBoxed.connectLongValue(mqttUri(TOPIC_BOXED_LONG));
    allBoxed.connectFloatValue(mqttUri(TOPIC_BOXED_FLOAT));
    allBoxed.connectDoubleValue(mqttUri(TOPIC_BOXED_DOUBLE));
    allBoxed.connectCharValue(mqttUri(TOPIC_BOXED_CHARACTER));

    allBoxed.connectBooleanValueTransformed(mqttUri(TOPIC_BOXED_BOOLEAN));
    allBoxed.connectIntValueTransformed(mqttUri(TOPIC_BOXED_INTEGER));
    allBoxed.connectShortValueTransformed(mqttUri(TOPIC_BOXED_SHORT));
    allBoxed.connectLongValueTransformed(mqttUri(TOPIC_BOXED_LONG));
    allBoxed.connectFloatValueTransformed(mqttUri(TOPIC_BOXED_FLOAT));
    allBoxed.connectDoubleValueTransformed(mqttUri(TOPIC_BOXED_DOUBLE));
    allBoxed.connectCharValueTransformed(mqttUri(TOPIC_BOXED_CHARACTER));
  }

  @Override
  protected void communicateSendInitialValue() {
    // empty
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    final boolean expectedBooleanValue = true;
    final int expectedIntValue = 1;
    final short expectedShortValue = 2;
    final long expectedLongValue = 3L;
    final float expectedFloatValue = 4.1f;
    final double expectedDoubleValue = 5.2;
    final char expectedCharValue = 'c';
    final String expectedStringValue = "6.3";

    publisher.publish(TOPIC_NATIVE_BOOLEAN, TestUtils.DefaultMappings.BoolToBytes(expectedBooleanValue));
    publisher.publish(TOPIC_NATIVE_INT, TestUtils.DefaultMappings.IntToBytes(expectedIntValue));
    publisher.publish(TOPIC_NATIVE_SHORT, TestUtils.DefaultMappings.ShortToBytes(expectedShortValue));
    publisher.publish(TOPIC_NATIVE_LONG, TestUtils.DefaultMappings.LongToBytes(expectedLongValue));
    publisher.publish(TOPIC_NATIVE_FLOAT, TestUtils.DefaultMappings.FloatToBytes(expectedFloatValue));
    publisher.publish(TOPIC_NATIVE_DOUBLE, TestUtils.DefaultMappings.DoubleToBytes(expectedDoubleValue));
    publisher.publish(TOPIC_NATIVE_CHAR, TestUtils.DefaultMappings.CharToBytes(expectedCharValue));
    publisher.publish(TOPIC_NATIVE_STRING, TestUtils.DefaultMappings.StringToBytes(expectedStringValue));

    publisher.publish(TOPIC_BOXED_BOOLEAN, TestUtils.DefaultMappings.BoolToBytes(expectedBooleanValue));
    publisher.publish(TOPIC_BOXED_INTEGER, TestUtils.DefaultMappings.IntToBytes(expectedIntValue));
    publisher.publish(TOPIC_BOXED_SHORT, TestUtils.DefaultMappings.ShortToBytes(expectedShortValue));
    publisher.publish(TOPIC_BOXED_LONG, TestUtils.DefaultMappings.LongToBytes(expectedLongValue));
    publisher.publish(TOPIC_BOXED_FLOAT, TestUtils.DefaultMappings.FloatToBytes(expectedFloatValue));
    publisher.publish(TOPIC_BOXED_DOUBLE, TestUtils.DefaultMappings.DoubleToBytes(expectedDoubleValue));
    publisher.publish(TOPIC_BOXED_CHARACTER, TestUtils.DefaultMappings.CharToBytes(expectedCharValue));

    TestUtils.waitForMqtt();

    assertEquals(expectedBooleanValue, integers.getBooleanValue());
    assertEquals(expectedIntValue, integers.getIntValue());
    assertEquals(expectedShortValue, integers.getShortValue());
    assertEquals(expectedLongValue, integers.getLongValue());
    assertEquals(expectedFloatValue, floats.getFloatValue(), TestUtils.DELTA);
    assertEquals(expectedDoubleValue, floats.getDoubleValue(), TestUtils.DELTA);
    assertEquals(expectedCharValue, chars.getCharValue());
    assertEquals(expectedStringValue, chars.getStringValue());

    assertEquals(expectedBooleanValue, integers.getBooleanValueTransformed());
    assertEquals(expectedIntValue, integers.getIntValueTransformed());
    assertEquals(expectedShortValue, integers.getShortValueTransformed());
    assertEquals(expectedLongValue, integers.getLongValueTransformed());
    assertEquals(expectedFloatValue, floats.getFloatValueTransformed(), TestUtils.DELTA);
    assertEquals(expectedDoubleValue, floats.getDoubleValueTransformed(), TestUtils.DELTA);
    assertEquals(expectedCharValue, chars.getCharValueTransformed());
    assertEquals(expectedStringValue, chars.getStringValueTransformed());

    assertEquals(expectedBooleanValue, allBoxed.getBooleanValue());
    assertEquals(expectedIntValue, allBoxed.getIntValue().intValue());
    assertEquals(expectedShortValue, allBoxed.getShortValue().shortValue());
    assertEquals(expectedLongValue, allBoxed.getLongValue().longValue());
    assertEquals(expectedFloatValue, allBoxed.getFloatValue(), TestUtils.DELTA);
    assertEquals(expectedDoubleValue, allBoxed.getDoubleValue(), TestUtils.DELTA);
    assertEquals(expectedCharValue, allBoxed.getCharValue().charValue());

    assertEquals(expectedBooleanValue, allBoxed.getBooleanValueTransformed());
    assertEquals(expectedIntValue, allBoxed.getIntValueTransformed().intValue());
    assertEquals(expectedShortValue, allBoxed.getShortValueTransformed().shortValue());
    assertEquals(expectedLongValue, allBoxed.getLongValueTransformed().longValue());
    assertEquals(expectedFloatValue, allBoxed.getFloatValueTransformed(), TestUtils.DELTA);
    assertEquals(expectedDoubleValue, allBoxed.getDoubleValueTransformed(), TestUtils.DELTA);
    assertEquals(expectedCharValue, allBoxed.getCharValueTransformed().charValue());
  }

  @Override
  public void closeConnections() {
    if (model != null) {
      model.ragconnectCloseConnections();
    }
  }

}
